#include "prechdr.hpp"
#include "tasocket.hpp"


TAsyncSocket::TAsyncSocket()
{
    // if this is not true, change the LPARAM value as
    //   necessary when posting a message to the notification
    //   window (it becomes "(LPARAM)this" for far models)
    ASSERT(sizeof(TAsyncSocket*) == sizeof(WORD));
    
    m_nIncomingBufferSize = 0;
    m_pIncomingBuffer     = NULL;
    m_pReadBuffer         = NULL;  // no outstanding read
    m_pWriteBuffer        = NULL;  // no outstanding write
    m_pWindowToNotify     = NULL;
    m_unConnectMessage    = 0;
    m_unClosedMessage     = 0;
    m_unErrorMessage      = 0;
}  // TAsyncSocket::TAsyncSocket


TAsyncSocket::~TAsyncSocket()
{
    ASSERT_VALID(this);
    Close();
}  // TAsyncSocket::~TAsyncSocket


void TAsyncSocket::SetNotification(CWnd* pWindowToNotify, UINT unConnectMessage, UINT unClosedMessage, UINT unErrorMessage)
{
    ASSERT(pWindowToNotify != NULL);
    ASSERT(unConnectMessage >= WM_USER);
    ASSERT(unClosedMessage >= WM_USER);
    ASSERT(unErrorMessage >= WM_USER);
    
    m_pWindowToNotify  = pWindowToNotify;
    m_unConnectMessage = unConnectMessage;
    m_unClosedMessage  = unClosedMessage;
    m_unErrorMessage   = unErrorMessage;

    ASSERT_VALID(this);
}  // TAsyncSocket::SetNotification


BOOL TAsyncSocket::IsReadOutstanding() const
{
    ASSERT_VALID(this);
    return(m_pReadBuffer != NULL);
}  // TAsyncSocket::IsReadOutstanding


BOOL TAsyncSocket::IsWriteOutstanding() const
{
    ASSERT_VALID(this);
    return(m_pWriteBuffer != NULL);
}  // TAsyncSocket::IsWriteOutstanding


BOOL TAsyncSocket::Create(UINT nSocketPort)
{
    ASSERT_VALID(this);
    return(CAsyncSocket::Create(nSocketPort,
                                SOCK_STREAM,
                                FD_CONNECT | FD_READ | FD_WRITE | FD_CLOSE,
                                NULL));
}  // TAsyncSocket::Create


BOOL TAsyncSocket::Initialize()
{
    ASSERT_VALID(this);
    return(AsyncSelect(FD_READ | FD_WRITE));
}  // TAsyncSocket::Initialize


void TAsyncSocket::Close()
{
    ASSERT_VALID(this);
    
    // save outstanding read/write information
    CompletionFn  pReadCompletionFn  = IsReadOutstanding() ? m_pReadCompletionFn : NULL;
    void*         pReadRefCon1       = m_pReadRefCon1;
    void*         pReadRefCon2       = m_pReadRefCon2;
    CompletionFn  pWriteCompletionFn = IsWriteOutstanding() ? m_pWriteCompletionFn : NULL;
    void*         pWriteRefCon1      = m_pWriteRefCon1;
    void*         pWriteRefCon2      = m_pWriteRefCon2;
    
    // reset internal state
    free(m_pIncomingBuffer);  // can free NULL pointer
    m_nIncomingBufferSize   = 0;
    m_pIncomingBuffer       = NULL;
    m_pReadBuffer           = NULL;  // no outstanding read
    m_pWriteBuffer          = NULL;  // no outstanding write

    CAsyncSocket::Close();

    // call any outstanding read request back with an aborted flag
    if(pReadCompletionFn != NULL)
    {
        (*pReadCompletionFn)(pReadRefCon1, pReadRefCon2, knResultAborted);
    }

    // call any outstanding write request back with an aborted flag
    if(pWriteCompletionFn != NULL)
    {
        (*pWriteCompletionFn)(pWriteRefCon1, pWriteRefCon2, knResultAborted);
    }
}  // TAsyncSocket::Close


BOOL TAsyncSocket::Read(void* lpBuf, int nBufLen, CompletionFn pfnCompletion, void* pRefCon1, void* pRefCon2)
{
    ASSERT_VALID(this);
    ASSERT(lpBuf != NULL);
    ASSERT(nBufLen > 0);
    ASSERT(pfnCompletion != NULL);
    
    if(IsReadOutstanding())
    {
        return(FALSE);  // only one outstanding read allowed
    }
    
    m_pReadBuffer       = lpBuf;
    m_nReadBufferLength = nBufLen;
    m_pReadCompletionFn = pfnCompletion;
    m_pReadRefCon1      = pRefCon1;
    m_pReadRefCon2      = pRefCon2;
    
    ProcessIncomingData();
    return(TRUE);
}  // TAsyncSocket::Read


BOOL TAsyncSocket::Write(const void* lpBuf, int nBufLen, CompletionFn pfnCompletion, void* pRefCon1, void* pRefCon2)
{
    ASSERT_VALID(this);
    ASSERT(lpBuf != NULL);
    ASSERT(nBufLen > 0);
    ASSERT(pfnCompletion != NULL);

    if(IsWriteOutstanding())
    {
        return(FALSE);  // only one outstanding write allowed
    }
    
    m_pWriteBuffer       = lpBuf;
    m_nWriteBufferLength = nBufLen;
    m_pWriteCompletionFn = pfnCompletion;
    m_pWriteRefCon1      = pRefCon1;
    m_pWriteRefCon2      = pRefCon2;
    
    ProcessOutgoingData();
    return(TRUE);
}  // TAsyncSocket::Write


void TAsyncSocket::ProcessIncomingData()
{
    ASSERT_VALID(this);
    
    // is there an outstanding read and enough data to complete it?
    if(m_pReadBuffer != NULL && m_nReadBufferLength <= m_nIncomingBufferSize)
    {
        // copy the data to the caller's buffer
        memcpy(m_pReadBuffer,
               m_pIncomingBuffer,
               m_nReadBufferLength);

        // "shift" the incoming buffer data forward
        m_nIncomingBufferSize -= m_nReadBufferLength;
        if(m_nIncomingBufferSize == 0)
        {
            free(m_pIncomingBuffer);
            m_pIncomingBuffer = NULL;
        }
        else
        {
            memmove(m_pIncomingBuffer,
                    m_pIncomingBuffer + m_nReadBufferLength,
                    m_nIncomingBufferSize);  // can move zero bytes
        }
        
        // call the read completion routine
        m_pReadBuffer = NULL;  // no more outstanding read
        (*m_pReadCompletionFn)(m_pReadRefCon1, m_pReadRefCon2, knResultOK);
    }
}  // TAsyncSocket::ProcessIncomingData


void TAsyncSocket::ProcessOutgoingData()
{
    ASSERT_VALID(this);

    // is there an outstanding write?
    if(m_pWriteBuffer != NULL)
    {
        int  nSent;
        
        nSent = Send(m_pWriteBuffer, m_nWriteBufferLength);
        if(nSent == SOCKET_ERROR)
        {
            if(GetLastError() == WSAEWOULDBLOCK)
            {
                // will try again later
            }
            else
            {
                // write failed, close and then tell the owner
                int  nLastError = GetLastError();
                Close();  // will destroy GetLastError()
                m_pWindowToNotify->PostMessage(m_unErrorMessage,
                                               (WPARAM)nLastError,
                                               MAKELPARAM((WORD)this, 0));
            }
        }
        else
        {
            ASSERT(nSent > 0);
            ASSERT(nSent <= m_nWriteBufferLength);
            
            m_pWriteBuffer = (void*)((char*)m_pWriteBuffer + nSent);
            m_nWriteBufferLength -= nSent;
            if(m_nWriteBufferLength == 0)
            {
                // call the write completion routine
                m_pWriteBuffer = NULL;
                (*m_pWriteCompletionFn)(m_pWriteRefCon1, m_pWriteRefCon2, knResultOK);
            }
        }
    }
}  // TAsyncSocket::ProcessOutgoingData


void TAsyncSocket::OnConnect(int nErrorCode)
{
    ASSERT_VALID(this);

    // close if an error occurred, and then tell the owner
    if(nErrorCode != 0)
    {
        Close();
    }
    m_pWindowToNotify->PostMessage(m_unConnectMessage,
                                   (WPARAM)nErrorCode,
                                   MAKELPARAM((WORD)this, 0));
}  // TAsyncSocket::OnConnect


void TAsyncSocket::OnReceive(int nErrorCode)
{
    ASSERT_VALID(this);

    DWORD  dwBytesAvailable;
    
    if(nErrorCode == 0 && IOCtl(FIONREAD, &dwBytesAvailable))
    {
        ASSERT(dwBytesAvailable > 0);
        ASSERT(dwBytesAvailable < INT_MAX);

        void*  pNewIncomingBuffer;
        
        pNewIncomingBuffer = realloc(m_pIncomingBuffer,
                                     m_nIncomingBufferSize + (int)dwBytesAvailable);  // can realloc NULL pointer
        if(pNewIncomingBuffer != NULL)
        {
            int  nRead;

            m_pIncomingBuffer = (char*)pNewIncomingBuffer;
            nRead = Receive(m_pIncomingBuffer + m_nIncomingBufferSize,
                            (int)dwBytesAvailable);
            if(nRead == SOCKET_ERROR)
            {
                // read failed, close and then tell the owner
                int  nLastError = GetLastError();
                Close();  // will destroy GetLastError()
                m_pWindowToNotify->PostMessage(m_unErrorMessage,
                                               (WPARAM)nLastError,
                                               MAKELPARAM((WORD)this, 0));
            }
            else if(nRead == 0)
            {
                // socket closed, close and then tell the owner
                Close();
                m_pWindowToNotify->PostMessage(m_unClosedMessage,
                                               (WPARAM)WSAECONNABORTED,
                                               MAKELPARAM((WORD)this, 0));
            }
            else
            {
                ASSERT(nRead > 0);
                ASSERT(nRead <= (int)dwBytesAvailable);
                
                m_nIncomingBufferSize += nRead;
                ProcessIncomingData();
            }
        }
        else
        {
            // out of memory, close and then tell the owner
            Close();
            m_pWindowToNotify->PostMessage(m_unErrorMessage,
                                           (WPARAM)WSAECONNABORTED,
                                           MAKELPARAM((WORD)this, 0));
        }
    }
    else
    {
        // socket error, close and then tell the owner
        Close();
        m_pWindowToNotify->PostMessage(m_unErrorMessage,
                                       (WPARAM)WSAECONNABORTED,
                                       MAKELPARAM((WORD)this, 0));
    }
}  // TAsyncSocket::OnReceive


void TAsyncSocket::OnSend(int nErrorCode)
{
    ASSERT_VALID(this);

    if(nErrorCode == 0)
    {
        ProcessOutgoingData();
    }
    else
    {
        // socket error, close and then tell the owner
        Close();
        m_pWindowToNotify->PostMessage(m_unErrorMessage,
                                       (WPARAM)WSAECONNABORTED,
                                       MAKELPARAM((WORD)this, 0));
    }
}  // TAsyncSocket::OnSend


void TAsyncSocket::OnClose(int nErrorCode)
{
    ASSERT_VALID(this);

    // other side closed socket, close and then tell the owner
    Close();
    m_pWindowToNotify->PostMessage(m_unClosedMessage,
                                   (WPARAM)WSAECONNABORTED,
                                   MAKELPARAM((WORD)this, 0));
}  // TAsyncSocket::OnClose


#ifdef _DEBUG
void TAsyncSocket::AssertValid() const
{
    CAsyncSocket::AssertValid();
    
    if(m_nIncomingBufferSize > 0)
    {
        ASSERT(m_pIncomingBuffer != NULL);
    }
    else
    {
        ASSERT(m_nIncomingBufferSize == 0);
        ASSERT(m_pIncomingBuffer == NULL);
    }
    
    if(m_pReadBuffer != NULL)
    {
        ASSERT(m_nReadBufferLength > 0);
        ASSERT(m_pReadCompletionFn != NULL);
    }
    
    if(m_pWriteBuffer != NULL)
    {
        ASSERT(m_nWriteBufferLength > 0);
        ASSERT(m_pWriteCompletionFn != NULL);
    }

    ASSERT(m_pWindowToNotify != NULL);
    ASSERT(m_unConnectMessage >= WM_USER);
    ASSERT(m_unClosedMessage >= WM_USER);
    ASSERT(m_unErrorMessage >= WM_USER);
}  // TAsyncSocket::AssertValid
#endif  // _DEBUG
